using System;
using System.Collections;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.ComponentModel;
using System.Data;

namespace GOST
{
    public class GOST2814789
    {
        
        private byte[] key = new byte[8];
        private BitArray baKey;
        
        /// <summary>
        /// 
        /// </summary>
        public string Key
        {
            get 
            {
                return this.ToStringFromByteArray(this.key);
            }
        }
        public byte[] KeyByte
        {
            get
            {
                return this.key;
            }
        }
        private List<byte> cryptMessage;

        /// <summary>
        ///   8  16   4 ,   0  15.
        /// </summary>
        private byte[,] matrixH = new byte[8, 16] {
										{4,10,9,2,13,8,0,14,6,11,1,12,7,15,5,3},
										{14,11,4,12,6,13,15,10,2,3,8,1,0,5,7,9},
										{5,8,1,13,10,3,4,2,14,15,12,7,6,0,9,11},
										{7,13,10,1,0,8,9,15,14,4,6,12,11,2,5,3},
										{6,12,7,1,5,15,13,8,4,10,9,14,0,3,11,2},
										{4,11,10,0,7,2,1,13,3,6,8,5,9,12,15,14},
										{13,11,4,1,3,15,5,9,0,10,14,7,6,8,2,12},
										{1,15,13,0,5,7,10,4,9,2,3,14,6,11,8,12} };
       

        public GOST2814789(byte[] key)
        {
            this.key = key;
            this.baKey = new BitArray(key);
        }

        public GOST2814789()
        {
            //  
            GenerateKey();
        }

        /// <summary>
        ///  
        /// </summary>
        /// <param name="source">  </param>
        /// <returns></returns>
        public byte[] Crypt(string source)
        {
            Encoding enc = System.Text.Encoding.GetEncoding("utf-8");
            //System.Text.UTF8Encoding enc = new System.Text.UTF8Encoding();
            return DoCrypt(enc.GetBytes(source));
        }

        /// <summary>
        ///  
        /// </summary>
        /// <param name="fs">Files strem</param>
        /// <returns>        crypt</returns>
        public void Crypt(FileStream fs)
        {
            BinaryWriter bw;

            BinaryReader br = new BinaryReader(fs);
            byte[] crypt = DoCrypt(br.ReadBytes((int)fs.Length));
            // save crypt file
            FileStream fsc = new FileStream(fs.Name+".crypt", FileMode.CreateNew);
            bw = new BinaryWriter(fsc);
            bw.Write(crypt);
            bw.Close();
            fsc.Close();
            fs.Close();
            // save key
            FileStream fsk = new FileStream(fs.Name + ".key", FileMode.CreateNew);
            bw = new BinaryWriter(fsk);
            bw.Write(this.key);
            bw.Close();
            fsk.Close();
        }

        /// <summary>
        ///  
        /// </summary>
        /// <param name="source">  </param>
        /// <returns> </returns>
        public string Decrypt(byte[] source)
        {
            //System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding(); //UTF8Encoding();
         //   return DoDecrypt(enc.GetBytes(source));
            return System.Text.UTF8Encoding.UTF8.GetString(DoDecrypt(source));
        }

        /// <summary>
        ///  
        /// </summary>
        /// <param name="fs">Files strem</param>
        /// <returns>        decrypt</returns>
        public void Decrypt(FileStream fs)
        {
            BinaryReader br;
            
            FileStream fsk = new FileStream(fs.Name.Replace(".crypt","") + ".key", FileMode.Open);
            br = new BinaryReader(fsk);
            byte[] key = br.ReadBytes((int)fs.Length);
            br.Close();
            fsk.Close();

            br = new BinaryReader(fs);
           // string dcrypt = System.Text.UTF8Encoding.UTF8.GetString(DoDecrypt(br.ReadBytes((int)fs.Length)));
            byte[] dcrypt = DoDecrypt(br.ReadBytes((int)fs.Length));
            FileStream fsc = new FileStream(fs.Name + ".decrypt", FileMode.CreateNew);
            BinaryWriter bw = new BinaryWriter(fsc);
            bw.Write(dcrypt);
            bw.Close();
            fsc.Close();
            fs.Close();
            

        }

        /// <summary>
        ///  /
        /// </summary>
        /// <param name="N">      64 =8</param>
        /// <param name="X">32    </param>
        /// <returns>  </returns>
        private byte[] MainStep(BitArray N, BitArray X)
        {
            byte[] bN = new byte[8];
            byte[] bN1 = new byte[4];
            byte[] bN2 = new byte[4];
            byte[] bKey = new byte[4];

            N.CopyTo(bN, 0);
            bN1[0] = bN[0];
            bN1[1] = bN[1];
            bN1[2] = bN[2];
            bN1[3] = bN[3];
            bN2[0] = bN[4];
            bN2[1] = bN[5];
            bN2[2] = bN[6];
            bN2[3] = bN[7];

            X.CopyTo(bKey, 0);

            uint N1 = BitConverter.ToUInt32(bN1, 0);
            uint N2 = BitConverter.ToUInt32(bN2, 0);
            uint iP = BitConverter.ToUInt32(bKey, 0);

            //  1.	  .         
            // 2^32      ,     ;
            uint sm = (uint)((N1 + iP) / (2 ^ 32));

            // 2.	 .  32- ,    , 
            //      4-  : S=(S0,S1,S2,S3,S4,S5,S6,S7).
            BitArray S = new BitArray(BitConverter.GetBytes(sm));
            
            //        
            byte[] bufer = new byte[8];
            for (int m = 0; m <= 7; m++)
            {
                //  4     S
                int smi = GetByteFromFourBit(S, m);
                //       H
                byte b = matrixH[m, smi];
                bufer[m] = b;
            }
            //    S   8 4 
            S = GetArray4FromByte(bufer);
            //  S   S1
            byte[] b1 = new byte[4];
            S.CopyTo(b1, 0);
            uint S1 = BitConverter.ToUInt32(b1, 0);

            //  3.	   11  
            S1 = S1 << 11;
            
            //  4.	 :  ,    3, 
            //     2     .
            //  XOR
            S1 = S1 ^ N2;

            // 5.	  :     
            //    ,         .
            N2 = N1;
            N1 = S1;

            List<byte> result = new List<byte>();
            result.AddRange(BitConverter.GetBytes(N1));
            result.AddRange(BitConverter.GetBytes(N2));
            return result.ToArray();
        }

        /// <summary>
        ///  
        /// </summary>
        /// <param name="N">64     </param>
        /// <returns>  </returns>
        private byte[] CryptStep(BitArray N)
        {
            BitArray blockKey;
            //    1..3 
            for (int k = 1; k <= 3; k++)
            {
                //     0..7
                for (int j = 0; j <= 7; j++)
                {
                    //   
                    blockKey = GetPartFromBitArray(this.baKey, j * 32, 32);
                    //   ,   
                    N = new BitArray(MainStep(N, blockKey));

                }
            }
            //     7..0
            for (int j = 7; j >= 0; j--)
            {
                blockKey = GetPartFromBitArray(this.baKey, j * 32, 32);
                N = new BitArray(MainStep(N, blockKey));
            }
            //   
            N = Exchange(N);
            byte[] res = new byte[8];
            N.CopyTo(res, 0);
            return res;
        }

        /// <summary>
        ///  
        /// </summary>
        /// <param name="N">64    </param>
        /// <returns>  </returns>
        private byte[] DecryptStep(BitArray N)
        {
            BitArray blockKey;
            //     0..7
            for (int j = 0; j <= 7; j++)
            {
                blockKey = GetPartFromBitArray(this.baKey, j * 32, 32);
                N = new BitArray(MainStep(N, blockKey));
            }

            //    1..3 
            for (int k = 1; k <= 3; k++)
            {
                //     7..0
                for (int j = 7; j >= 0; j--)
                {
                    //   
                    blockKey = GetPartFromBitArray(this.baKey, j * 32, 32);
                    //   ,   
                    N = new BitArray(MainStep(N, blockKey));
                }
            }
            //   
            N = Exchange(N);
            byte[] res = new byte[8];
            N.CopyTo(res, 0);
            return res;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="source">    </param>
        /// <returns> </returns>
        private byte[] DoCrypt(byte[] source)
        {
            List<byte> lbMes = new List<byte>(source);
            int countadd = ((((lbMes.Count * 8) / 64) * 64 + 64) - lbMes.Count * 8) / 8;
            if (countadd > 0)
            {
                for (int add = 0; add < countadd; add++)
                {
                    //    
                    lbMes.Add(0);
                }
            }

            //    
            BitArray Message = new BitArray(lbMes.ToArray());
            //  
            List<byte> cryptmes = new List<byte>();
            //   
            for (int partmes = 0; partmes < Message.Count / 64; partmes++)
            {
                //      64 
                BitArray blockMes = GetPartFromBitArray(Message, partmes * 64, 64);
                //  
                cryptmes.AddRange(CryptStep(blockMes));
            }

            return cryptmes.ToArray();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="source">    </param>
        /// <returns> </returns>
        private byte[] DoDecrypt(byte[] source)
        {
            //     
            BitArray CryptMessage = new BitArray(source);
            //   
            List<byte> decryptmes = new List<byte>();
            //   
            for (int partmes = 0; partmes < CryptMessage.Count / 64; partmes++)
            {
                BitArray blockMes = GetPartFromBitArray(CryptMessage, partmes * 64, 64);
                decryptmes.AddRange(DecryptStep(blockMes));
            }
            return decryptmes.ToArray();
        }

        /// <summary>
        ///  
        /// </summary>
        public void GenerateKey()
        {
            List<byte> lkey = new List<byte>();
            //  8    
            for (int i = 0; i <= 7; i++)
            {
                Random rndgen = new Random();
                uint k = (uint)rndgen.Next(int.MaxValue);
                lkey.AddRange(BitConverter.GetBytes(k));
                System.Threading.Thread.Sleep(10);
            }
            //         
            this.key = lkey.ToArray();
            this.baKey = new BitArray(this.key);
        }


   

       /// <summary>
       ///     
       /// </summary>
       /// <param name="val"> </param>
       /// <returns></returns>
        public string ToStringFromByteArray(byte[] val)
        {
            char[] chars = new char[val.Length];
            string result = "";
            for (int i = 0; i < val.Length; i++)
            {
                chars[i] = Convert.ToChar(val[i]);
                result += chars[i];
            }
            return result;
        }

        /// <summary>
        ///     0  1
        /// </summary>
        /// <param name="val"> 0  1</param>
        /// <returns></returns>
        private int ToIntFromString(string val)
        {
            val = val.Replace("False", "0");
            val = val.Replace("True", "1");
            char[] chars = val.ToCharArray();
            byte[] result = new byte[chars.Length];
            return Convert.ToInt32(val, 2);

        }

        /// <summary>
        ///    BitArray
        /// </summary>
        /// <param name="source">BitArray </param>
        /// <param name="start"> </param>
        /// <param name="lenght"></param>
        /// <returns>BitArray </returns>
        private BitArray GetPartFromBitArray(BitArray source, int start, int lenght)
        {
            BitArray result = new BitArray(lenght);

            int idx = 0;
            for (int i = start; i < start + lenght; i++)
            {
                result[idx] = source[i];
                idx++;
            }
            return result;
        }

        /// <summary>
        ///   
        /// </summary>
        /// <param name="source"> </param>
        /// <returns>  </returns>
        private BitArray Exchange(BitArray source)
        {
            BitArray buf1 = new BitArray(source.Length / 2);
            BitArray buf2 = new BitArray(source.Length / 2);
            for (int i = 0; i < source.Length / 2; i++)
            {
                buf1[i] = source[i];
            }
            int idx = 0;
            for (int i = source.Length / 2; i < source.Length; i++)
            {
                buf2[idx] = source[i];
                idx++;
            }

            for (int i = 0; i < source.Length / 2; i++)
            {
                source[i] = buf2[i];
            }
            idx = 0;
            for (int i = source.Length / 2; i < source.Length; i++)
            {
                source[i] = buf1[idx];
                idx++;
            }


            return source;
        }

        
        private int GetByteFromFourBit(BitArray source, int part)
        {
            int idx1 = part * 4;
            int idx2 = part * 4 + 1;
            int idx3 = part * 4 + 2;
            int idx4 = part * 4 + 3;

            string res = source[idx1].ToString() + source[idx2].ToString() + source[idx3].ToString() + source[idx4].ToString();
            return ToIntFromString(res);
        }

        private BitArray GetArray4FromByte(byte[] source)
        {
            BitArray ba = new BitArray(32);

            string res = "";
            for (int m = 0; m < 8; m++)
            {
                string s = Convert.ToString(source[m], 2);
                switch (s.Length)
                {
                    case 0: s = "0000";
                        break;
                    case 1: s = "000" + s;
                        break;

                    case 2: s = "00" + s;
                        break;

                    case 3: s = "0" + s;
                        break;
                }
                res += s;
            }
            for (int i = 0; i < 32; i++)
            {
                if (res[i] == '1')
                {

                    ba[i] = true;
                }
                else
                {
                    ba[i] = false;
                }
            }
            return ba;
        }

    }
}
